/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*  Spruce
 *  Copyright (C) 1999-2000 Jeffrey Stedfast
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#ifdef HAVE_CONFIG_H
#	include <config.h>
#endif

#include "mesg.h"
#include "parse.h"

guint mesg_get_header_len (const gchar *mesg)
{
	gchar *start_of_body;
	g_return_val_if_fail (mesg, 0);

	/* RFC822 compliant message headers are separated from the 
	   message body by CRLF */
	start_of_body = strstr (mesg, "\n\n");
	g_return_val_if_fail (start_of_body, 0);

	return (guint)(start_of_body - mesg);
}

gchar *mesg_get_trimmed_header (const gchar *message, gint trim_level, 
			        guint hdrlen)
{
	/* First entry in showfields[] should not be preceeded by a newline,
	since data passed through this function twice will have the same order
	and the first field will not be preceded by a newline. */
	static char *showfields[] = { "Date:", "\nTo:", "\nFrom:", "\nCC:", 
				      "\nBCC:", "\nReply-To:", "\nSender:", 
				      "\nSubject:", "\0" };

	gchar *hdrtrimmed, *header;
	gchar *fieldvalue, *fieldname;
	gint i;
	
	/* Return no headers (HEADER_TRIMLEVEL_KEEP_NONE) */
	if (trim_level == HEADER_TRIMLEVEL_KEEP_NONE)
		return NULL;
	
	/* Return all headers (HEADER_TRIMLEVEL_KEEP_ALL) */
	if (trim_level == HEADER_TRIMLEVEL_KEEP_ALL)
		return g_strndup (message, hdrlen);
	
	
	/* Otherwise, return partial headers (HEADER_TRIMLEVEL_KEEP_PARTIAL)
	   by reconstructing a clean set of headers the values we want */
	
	/* Copy header plus two trailing \n's */
	header = g_strndup (message, hdrlen + 2);
	
	/* Allocate memory for size of known header + 3 \n's + NULL */
	hdrtrimmed = g_malloc0 (hdrlen + 4);
	i = 0;
	
	while (*showfields[i] != '\0') {
		fieldvalue = mesg_get_header_field (header, showfields[i]);
		if (fieldvalue != NULL) {
			g_strstrip (fieldvalue);
			fieldname = showfields[i];
			if (*fieldname == '\n')
				fieldname++;
			strcat (hdrtrimmed, fieldname);
			strcat (hdrtrimmed, " ");
			strcat (hdrtrimmed, fieldvalue);
			if (*showfields[i+1] != '\0')
				strcat (hdrtrimmed, "\n");
		}
		i++;
	}
	
	g_free (header);
	
	return hdrtrimmed;
}

gchar *mesg_trim_header (const gchar *message, const gint trim_level)
{
	gchar *mesg;
	gchar *start, *ptr, *ptr_mesg;
	gchar *hdrtrimmed;
	guint  hdrlen;
	
	mesg = g_strdup (message);
	
	start = mesg;
	hdrlen = mesg_get_header_len (mesg);
	if (trim_level != HEADER_TRIMLEVEL_KEEP_ALL)
		start += hdrlen;
	
	if (trim_level == HEADER_TRIMLEVEL_KEEP_PARTIAL) {
		/* show partial header only */
		hdrtrimmed = mesg_get_trimmed_header (mesg, trim_level, hdrlen);
		start -= strlen (hdrtrimmed);
		strncpy (start, hdrtrimmed, strlen (hdrtrimmed));
		g_free (hdrtrimmed);
	}
	
	/* advance to the actual beginning of the message (gets rid of any
	 * extra line feeds that might be between the header and the actual
	 * message). */
	for ( ; *start != '\0'; start++)
		if (*start != '\n')
			break;
	
	for (ptr_mesg = mesg, ptr = start; *ptr != '\0'; ptr_mesg++, ptr++)
		*ptr_mesg = *ptr;
	
	*ptr_mesg = '\0';
	
	return mesg;
}

gchar *mesg_get_header_field (gchar *message, gchar *field)
{
	static gchar buffer[512];
	gchar *header, *index, temp_line[512];
	gchar *mesg_header; 
	gint i, done, first;
	
	/* lets find the end of the header info */
	index = strstr (message, "\n\n");
	if (index == NULL)
		return (gchar *)NULL;
	
	index += 2;
	
	/* now lets make a copy of the header info */
	mesg_header = g_malloc0 (index - message + 1);
	strncpy (mesg_header, message, index - message);
	mesg_header[index - message] = '\0';
	
	index = strstrcase (mesg_header, field);
	if (index == NULL) {
		g_free (mesg_header);
		return NULL;
	}
	
	/* NULL out our buffer */
	memset (buffer, 0, sizeof (buffer));
	
	/* advance to the start of the header field */
	index += strlen (field) + 1;
	
	done = 0;
	first = 1;
	/* we'll get data until the first char in the line (except the first 
	   line) is not a tab or a space (indicates a new header field) */
	while (done == 0) {
		i = 0;
		while ((*index != '\n') && (i < sizeof (temp_line) - 1)) {
			temp_line[i] = *index;
			i++;
			index++;
		}
		temp_line[i] = '\0';
		
		if (!first) {
			if ((temp_line[0] != '\t') && (temp_line[0] != ' '))
				done = 1;
		}
		
		if (!done) {
			i = 0;
			while (isspace (temp_line[i]))
				i++;
			strcat (buffer, " ");
			strncat (buffer, temp_line + i, sizeof (buffer)-1);
			index++;
			first = 0;
		}
	}
	
	g_strstrip (buffer);
	parse_8bit (buffer);
	
	header = buffer;
	
	g_free (mesg_header);
	
	return header;
}

gchar *mesg_strip_header (gchar *message)
{
	gchar *hdrstripped;
	guint hdrlen;

        hdrlen = mesg_get_header_len (message);
	hdrstripped = g_strdup(message);
	strcut (hdrstripped, 0, hdrlen + 2);

	return hdrstripped;
}
