/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*  Spruce
 *  Copyright (C) 1999-2000 Jeffrey Stedfast
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#ifdef HAVE_CONFIG_H
#	include <config.h>
#endif

#include "smtpio.h"

#define d(x) x

static gchar *smtp_errlist[] = { "",
				 ": Hostname failed to resolve", 
				 ": Socket failed to open",
				 ": No route to host?",
				 ": Server failed to acknowledge",
				 ": HELO send()",
				 ": HELO recv()",
				 ": MAIL FROM send()",
				 ": MAIL FROM recv()",
				 ": RCPT TO send()",
				 ": RCPT TO recv",
				 ": DATA send()",
				 ": DATA recv()",
				 ": BODY recv()",
				 ": RSET send()",
				 ": RSET recv()",
				 ": QUIT send()",
				 ": QUIT recv()",
				 ": Sender invalid",
				 ": Recipient invalid"
};

extern gint errno;
extern gint timeout;
SMTPError smtperrno;

static gchar smtpErrMesg[512];
static gboolean smtp_is_esmtp = FALSE;
static GList *esmtp_supported_authtypes = NULL;

static GList *esmtp_get_authtypes (gchar *buffer);

void smtperror (gchar *reason, gint len)
{
	strncpy (reason, smtp_errlist[smtperrno], len);
	if (strlen (reason) + strlen (smtpErrMesg) + 3 < len) {
		strcat (reason, ": ");
		strcat (reason, smtpErrMesg);
	}
}

void email_destroy (Email *msg)
{
	g_free (msg->from);
	msg->from = NULL;
	g_free (msg->from_addr);
	msg->from_addr = NULL;
	g_free (msg->to);
	msg->to = NULL;
	g_free (msg->to_addr);
	msg->to_addr = NULL;
	g_free (msg->cc);
	msg->cc = NULL;
	g_free (msg->cc_addr);
	msg->cc_addr = NULL;
	g_free (msg->bcc);
	msg->bcc = NULL;
	g_free (msg->bcc_addr);
	msg->bcc_addr = NULL;
	g_free (msg->replyto);
	msg->replyto = NULL;
	g_free (msg->header);
	msg->header = NULL;
	g_free (msg->subject);
	msg->subject = NULL;
	g_free (msg->xmailer);
	msg->xmailer = NULL;
	g_free (msg->in_reply_to);
	msg->in_reply_to = NULL;
	g_free (msg->references);
	msg->references = NULL;
	g_free (msg->mime_version);
	msg->mime_version = NULL;
	g_free (msg->content_type);
	msg->content_type = NULL;
	g_free (msg->content_transfer_encoding);
	msg->content_transfer_encoding = NULL;
	g_free (msg->body);
	msg->body = NULL;
}

gint format_date (gchar *date, size_t date_size)
{
	time_t t;
	gint ret;
	gchar save_locale[32], *locale;
	
	locale = setlocale (LC_TIME, "");
	if (locale != NULL) {
		strncpy (save_locale, locale, 31);
	}
	setlocale (LC_TIME, "POSIX");
	
	time (&t);
	
	memset (date, 0, date_size);
	/* if (strftime (date, date_size, "%a, %d %b %Y %H:%M:%S GMT", gmtime (&t)) == 0) */
	ret = strftime (date, date_size-1, "%a, %d %b %Y %H:%M:%S %z", localtime (&t));
	
	if (locale != NULL)
		setlocale (LC_TIME, save_locale);
   	
	/* No characters written to date string. */
	if (ret == 0)
		return ERROR;
	else
		return SUCCESS;
}

gint format_header (Email *msg)
{
	/* format the header information */
	GPtrArray *hdrs;
	gchar date[65], *hdr_basic;
	static Version smtpio_ver;
	
	smtpio_ver.part.major = VER_MAJOR;
	smtpio_ver.part.delim1 = '.';
	smtpio_ver.part.minor = VER_MINOR;
	smtpio_ver.part.delim2 = '.';
	smtpio_ver.part.release = VER_REL;
	smtpio_ver.part.null = '\0';
	
	/* Get the date/time. */
	format_date (date, sizeof (date));
	
	/* Create the header. */
	hdr_basic = g_strdup_printf ("Date: %s\r\nFrom: %s\r\nTo: %s", date, msg->from, msg->to);
	
	hdrs = g_ptr_array_new ();
	g_ptr_array_add (hdrs, hdr_basic);
	
	if (msg->cc && *msg->cc)
		 g_ptr_array_add (hdrs, g_strdup_printf ("CC: %s", msg->cc));
	
	g_ptr_array_add (hdrs, g_strdup_printf ("Subject: %s", msg->subject ? msg->subject : "No Subject"));
	
	if (msg->replyto)
		g_ptr_array_add (hdrs, g_strdup_printf ("Reply-To: %s", msg->replyto));
	
	if (msg->in_reply_to)
		g_ptr_array_add (hdrs, g_strdup_printf ("In-Reply-To: %s", msg->in_reply_to));
	
	if (msg->references)
		g_ptr_array_add (hdrs, g_strdup_printf ("References: %s", msg->references));
	
	g_ptr_array_add (hdrs, g_strdup_printf ("X-Mailer: %s w/smtpio %s", msg->xmailer, smtpio_ver.string));
	
	if (msg->mime_version)
		 g_ptr_array_add (hdrs, g_strdup_printf ("MIME-Version: %s", msg->mime_version));
	
	if (msg->content_type)
		 g_ptr_array_add (hdrs, g_strdup_printf ("Content-Type: %s", msg->content_type));
	
	if (msg->content_transfer_encoding)
		g_ptr_array_add (hdrs, g_strdup_printf ("Content-Transfer-Encoding: %s",
							msg->content_transfer_encoding));
	
	msg->header = g_strjoinv ("\r\n", (char **)hdrs->pdata);
	
	g_ptr_array_free (hdrs, TRUE);
	
	return SUCCESS;
}

static GList *esmtp_get_authtypes (gchar *buffer)
{
	GList *ret = NULL;
	gchar *start, *end;
	
	/* advance to the first token */
	for (start = buffer; *start == ' ' || *start == '='; start++);
	
	for ( ; *start; ) {
		/* advance to the end of the token */
		for (end = start; *end && *end != ' '; end++);
		
		ret = g_list_append (ret, g_strndup (start, end - start));
		
		/* advance to the next token */
		for (start = end; *start == ' '; start++);
	}
	
	return ret;
}

gint get_email_addr_from_text (gchar *addr, gchar *text)
{
	/* get the actual email address from the string passed and place it in *addr
	 * we can assume the address will be in one of the following forms:
	 * 1) The Name <person@host.com>
	 * 2) <person@host.com>
	 * 3) person@host.com
	 * 4) person@host.com (The Name)
	 */
	
	gchar *addr_strt;         /* points to start of addr */
	gchar *addr_end;          /* points to end of addr */
	gchar *ptr1, *ptr2;
	
	
	/* check the incoming args */
	if (!addr || !text || *text=='\0')
		return ERROR;
	
	/* scan the string for an open brace */
	for (addr_strt = text; *addr_strt; addr_strt++) 
		if (*addr_strt == '<')
			break;
	
	if (*addr_strt) {
		/* we found an open brace */
		/* let's look for it's counterpart */
		for (addr_end = addr_strt; *addr_end; addr_end++)
			if (*addr_end == '>') 
				break;
		
		/* if we didn't find it, or braces are empty... */
		if (!(*addr_end) || (addr_strt == addr_end - 1))
			return ERROR;
		
		/* addr_strt points to '<' and addr_end points to '>'.
		 * Now let's adjust 'em slightly to point to the beginning
		 * and ending of the email addy
		 */
		addr_strt++;
		addr_end--;
	} else {
		/* no open brace...assume type 3 or 4? */
		addr_strt = text;
			
		/* find the end of the email addr/string */
		for (addr_end = addr_strt; *addr_end || *addr_end == ' '; addr_end++);
		
		addr_end--;       /* points to NULL, move it back one char */
	}
	
	/* now addr_strt & addr_end point to the beginning & ending of the email addy */
	
	/* copy the string into addr */
	for (ptr1 = addr_strt, ptr2 = addr; ptr1 <= addr_end; ptr1++, ptr2++)
		*ptr2 = *ptr1;
	
	for (ptr1 = addr_strt; ptr1 <= addr_end; ptr1++)    /* look for an '@' sign */
		if (*ptr1 == '@')
			break;
	
	if (*ptr1 != '@') {
		/* here we found out the name doesn't have an '@' part
		 * let's figure out what machine we're on & stick it on the end
		 * woops.. forgot the '@' sign, gotta stick that in the reply
		 * ptr2 should still point to the next place to copy to in addr
		 */
		
		*ptr2++ = '@';
		if (gethostname (ptr2, MAXHOSTNAMELEN))
			return ERROR;
	} else                  /* there's an @ sign already in the email addy */
		*ptr2 = '\0';   /* add a null to the end of the string */
	
	return SUCCESS;
}

gint helo (gint sock)
{
	/* say hello to the server */
	gchar buffer[512];
	gchar *cmdbuf;
	struct sockaddr_in localaddr;
	guint32 addrlen;
	struct hostent *host;
	
	/* get the localhost name */
	addrlen = sizeof (localaddr);
	getsockname (sock, (struct sockaddr*)&localaddr, &addrlen);
	host = gethostbyaddr ((char *)&localaddr.sin_addr, sizeof (localaddr.sin_addr), AF_INET);
	
	/* hiya server! how are you today? */
	if (smtp_is_esmtp) {
		if (host && host->h_name)
			cmdbuf = g_strdup_printf ("EHLO %s\r\n", host->h_name);
		else
			cmdbuf = g_strdup_printf ("EHLO [%s]\r\n", inet_ntoa (localaddr.sin_addr));
	} else {
		if (host && host->h_name)
			cmdbuf = g_strdup_printf ("HELO %s\r\n", host->h_name);
		else
			cmdbuf = g_strdup_printf ("HELO [%s]\r\n", inet_ntoa (localaddr.sin_addr));
	}
	
	if (send (sock, cmdbuf, strlen (cmdbuf), 0) < 0) {
		g_free (cmdbuf);
		smtperrno = S_ERR_HELO_SEND;
		strncpy (smtpErrMesg, "request timed out.", sizeof (smtpErrMesg)-1);
		
		return ERROR;
	}
	g_free (cmdbuf);
	
	do {
		/* Check for "250" */
		memset (buffer, 0, sizeof (buffer));
		if ((recvline (sock, buffer, sizeof (buffer)-1) < 0) || strncmp (buffer, "250", 3)) {
			smtperrno = S_ERR_HELO_RECV;
			strncpy (smtpErrMesg, buffer, sizeof (smtpErrMesg)-1);
			
			return ERROR;
		}
		if (smtp_is_esmtp && strstr (buffer, "AUTH")) {
			/* parse for supported AUTH types */
			char *auths = strstr (buffer, "AUTH") + 4;
			
			esmtp_supported_authtypes = esmtp_get_authtypes (auths);
		}
	} while (*(buffer+3) == '-'); /* if we got "250-" then loop again */
	
	return SUCCESS;
}

gint mail (gint sock, gchar *sender)
{
	/* we gotta tell the smtp server who we are. (our email addy) */
	gchar buffer[512];
	gchar *cmdbuf;
	
	cmdbuf = g_strdup_printf ("MAIL FROM: <%s>\r\n", sender);
	if (send (sock, cmdbuf, strlen (cmdbuf), 0) < 0) {
		g_free (cmdbuf);
		smtperrno = S_ERR_MAIL_SEND;
		strcpy (smtpErrMesg, "request timed out.");
		
		return ERROR;
	}
	g_free (cmdbuf);
	
	do {
		/* Check for "250 Sender OK..." */
		memset (buffer, 0, sizeof (buffer));
		if ((recvline (sock, buffer, sizeof (buffer)-1) < 0) || strncmp (buffer, "250", 3)) {
			smtperrno = S_ERR_MAIL_RECV;
			strncpy (smtpErrMesg, buffer, sizeof (smtpErrMesg)-1);
			
			return ERROR;
		}
	} while (*(buffer+3) == '-'); /* if we got "250-" then loop again */
	
	return SUCCESS;
}

gint rcpt (gint sock, gchar *recipient, gchar *orig_rcpt)
{
	/* we gotta tell the smtp server who we are going to be sending
	 * our email to */
	gchar buffer[512];
	gchar *cmdbuf;

#undef I_KNOW_WHAT_ORCPT_IS_FOR
#ifdef I_KNOW_WHAT_ORCPT_IS_FOR
	if (!orig_rcpt || strcmp (recipient, orig_rcpt) == 0)
		/* No original recipient specified or same as recipient. */ 
		cmdbuf = g_strdup_printf ("RCPT TO: <%s>\r\n", recipient);
	else
		cmdbuf = g_strdup_printf ("RCPT TO: <%s> ORCPT=<%s>\r\n", recipient, orig_rcpt);
#else
	cmdbuf = g_strdup_printf ("RCPT TO: <%s>\r\n", recipient);
#endif
	
	if (send (sock, cmdbuf, strlen (cmdbuf), 0) < 0) {
		g_free (cmdbuf);
		smtperrno = S_ERR_RCPT_SEND;
		strncpy (smtpErrMesg, "request timed out.", sizeof (smtpErrMesg)-1);
		return ERROR;
	}
	g_free (cmdbuf);
	
	do {
		/* Check for "250 user@host.com... Recipient ok" */
		memset (buffer, 0, sizeof (buffer));
		if ((recvline (sock, buffer, sizeof (buffer)-1) < 0) || strncmp (buffer, "250", 3)) {
			smtperrno = S_ERR_RCPT_RECV;
			strncpy (smtpErrMesg, buffer, sizeof (smtpErrMesg)-1);
			
			return ERROR;
		}
	} while (*(buffer+3) == '-'); /* if we got "250-" then loop again */
	
	return SUCCESS;
}

gint data (gint sock, gchar *header, gchar *body)
{
	/* now we can actually send what's important :p */
	gchar buffer[512];
	gchar *cmdbuf;
	gchar data[256];
	gint i, y, len;
	
	cmdbuf = g_strdup ("DATA\r\n");
	if (send (sock, cmdbuf, strlen (cmdbuf), 0) < 0) {
		g_free (cmdbuf);
		smtperrno = S_ERR_DATA_SEND;
		strncpy (smtpErrMesg, "request timed out.", sizeof (smtpErrMesg)-1);
		return ERROR;
	}
	g_free (cmdbuf);
	
	if ( (recvline (sock, buffer, sizeof (buffer)-1) < 0)
	     || (find_string (buffer, StrtMailInpt) == -1) ) {
		/* we should have gotten instructions on how to use the DATA command:
		 * 354 Enter mail, end with "." on a line by itself
		 */
		smtperrno = S_ERR_DATA_RECV;
		strncpy (smtpErrMesg, buffer, sizeof (smtpErrMesg)-1);
		return ERROR;
	}
	
	/* now to send the actual data */
	/* send the header information */
	send (sock, header, strlen (header), 0);
	
	send (sock, "\r\n\r\n", 4, 0);
	
	memset (data, 0, sizeof (data));
	
	len = strlen (body);
	
	/* send the body of the message */
	for (i = 0; i < len; ) {
		data[0] = '\0';
		y = 0;
		while (y < 200 && i < len) {
			if (body[i] == '.' && body[i-1] == '\n' && body[i+1] == '\n') {
				/* if we have a lone period on a line... */
				data[y] = '.';
				data[y+1] = '.';
				y += 2;
				i++;
			} else {
				if (body[i] == '\n' && body[i-1] != '\r') {
					data[y] = '\r';
					data[y+1] = '\n';
					y += 2;
					i++;
				} else {
					data[y] = body[i];
					y++;
					i++;
				}
			}
		}
		data[y] = '\0';
		send (sock, data, strlen (data), 0);
	}
	
	/* terminate the message body */
	send (sock, "\r\n.\r\n", 5, 0);	

	do {
		/* Check for "250 Mail accepted" */
		memset (buffer, 0, sizeof (buffer));
		if ((recvline (sock, buffer, sizeof (buffer)-1) < 0) || strncmp (buffer, "250", 3)) {
			smtperrno = S_ERR_BODY_RECV;
			strncpy (smtpErrMesg, buffer, sizeof (smtpErrMesg)-1);
			
			return ERROR;
		}
	} while (*(buffer+3) == '-'); /* if we got "250-" then loop again */
	
	return SUCCESS;
}

gint rset (gint sock)
{
	/* we are going to reset the smtp server (just to be nice) */
	gchar buffer[512];
	gchar *cmdbuf;
	
	cmdbuf = g_strdup ("RSET\r\n");
	if (send (sock, cmdbuf, strlen (cmdbuf), 0) < 0) {
		g_free (cmdbuf);
		smtperrno = S_ERR_RSET_SEND;
		strncpy (smtpErrMesg, "request timed out.", sizeof (smtpErrMesg)-1);
		return ERROR;
	}
	g_free (cmdbuf);
	
	do {
		/* Check for "250 OK" */
		memset (buffer, 0, sizeof (buffer));
		if ((recvline (sock, buffer, sizeof (buffer)-1) < 0 ) || strncmp (buffer, "250", 3 )) {
			smtperrno = S_ERR_RSET_RECV;
			strncpy (smtpErrMesg, buffer, sizeof (smtpErrMesg)-1);
			
			return ERROR;
		}
	} while (*(buffer+3) == '-'); /* if we got "250-" then loop again */
	
	return SUCCESS;
}

gint quit (gint sock)
{
	/* lets tell the server we want to quit now */
	gchar buffer[512];
	gchar *cmdbuf;
	
	cmdbuf = g_strdup ("QUIT\r\n");
	if (send (sock, cmdbuf, strlen (cmdbuf), 0) < 0) {
		g_free (cmdbuf);
		smtperrno = S_ERR_QUIT_SEND;
		strncpy (smtpErrMesg, "request timed out.", sizeof (smtpErrMesg)-1);
		
		return ERROR;
	}
	g_free (cmdbuf);
	
	do {
		/* Check for "221 Service closed" */
		memset (buffer, 0, sizeof (buffer));
		if ((recvline (sock, buffer, sizeof (buffer)-1) < 0) || strncmp (buffer, "221", 3)) {
			/* mr server didn't say bye-bye :( */
			smtperrno = S_ERR_QUIT_RECV;
			strncpy (smtpErrMesg, buffer, sizeof (smtpErrMesg)-1);
			return ERROR;
		}
	} while (*(buffer+3) == '-'); /* if we got "221-" then loop again */
	
	return SUCCESS;
}

gint send_message (Server *server, Email *msg)
{
	/* this will interface with a smtp daemon on a foreign host
	 * and send an email message. You should call format_header()
	 * previous to calling this function. You may also want to call
	 * email_addr_from_text () to get the actual address. The variables sender[]
	 * and recipient[] must be in the format: user@host.com
	 * which is the format that email_addr_from_text () will return.
	 * "server" is of type SmtpHost and contains a hostname, ip, and port
	 * number.
	 */
	
	gint sock;
	gint fields, i, num;
	gchar buffer[512];
	gchar *buff = NULL;
	
	/* make sure we aren't claiming an error */
	smtperrno = S_NO_ERROR;
	
	/* null the error string */
	memset (smtpErrMesg, 0, sizeof (smtpErrMesg));
	
	/* set the bits to zero (yeah, I know...kinda complex 
	 * to do it this way but it's fun??) then set it to DELIVERING
	 * Tell me again why we don't just go: msg->status = DELIVERING; ?
	 */
	msg->status = (msg->status & 0) | DELIVERING;
	
	/* resolve the hostname */
	if (!Resolve (server)) {
		smtperrno = S_ERR_RESOLVE;
		return ERROR;
	}
	
	/* assign and open a socket */
	sock = socket (AF_INET, SOCK_STREAM, IPPROTO_TCP);
	if (sock < 0) {
		/* socket failed to open */
		smtperrno = S_ERR_SOCKET;
		return ERROR;
	}
	
	/* lets set up information about the server we will be connecting to... */
	server->sin.sin_family = AF_INET;
	server->sin.sin_port = htons (server->port);
	
	/* call the server */
	if (connect_timeo (sock, (struct sockaddr*)&server->sin, sizeof (server->sin), timeout) < 0) {
		/* we failed to connect */
		smtperrno = S_ERR_CONNECT;
		close (sock);
		return ERROR;
	}
	
	do  {
		/* Check for 220 Service Ready */
		memset (buffer, 0, sizeof (buffer));
		if ((recvline (sock, buffer, sizeof (buffer)-1) < 0 ) || strncmp (buffer, "220", 3)) {
			smtperrno = S_ERR_SERV_ACK;
			close (sock);
			return ERROR;
		}
		if (strstr (buffer, "ESMTP"))
			smtp_is_esmtp = TRUE;
	} while (*(buffer+3) == '-'); /* keep looping until we don't get "220-" */
	
	/* send the email */
	
	if (!helo (sock)) {
		close (sock);
		return ERROR;
	}
	
	if (esmtp_supported_authtypes != NULL) {
		/* not really supported yet, but we can at least show what auth types are supported */
		d(fprintf (stderr, "smtpio: %s requires AUTH\n", server->hostname));
		num = g_list_length (esmtp_supported_authtypes);
		for (i = 0; i < num; i++)
			d(fprintf (stderr, "Supported AUTH: %s\n",
				   (char *) g_list_nth_data (esmtp_supported_authtypes, i)));
		g_list_free (esmtp_supported_authtypes);
		esmtp_supported_authtypes = NULL;
	}
	
	if (!mail (sock, msg->from_addr)) {
		close (sock);
		return ERROR;
	}
	
	/* Determine number of direct recipients.  Do not CC or BCC if no direct 
	 * recipients are specified. */
	fields = get_num_fields (msg->to_addr, ',');
	
	if (fields > 0) {
		/* We will consider the first (or only) recipient to be the original recipient. */
		gchar *orig_rcpt = get_field (msg->to_addr, 1, ',');
		
		/* Do a RCPT TO for each of the direct recipients. */
		for (i = 1; i <= fields; i++) {
			if (i != 1)
				buff = get_field (msg->to_addr, i, ',');
			else
				buff = g_strdup (orig_rcpt);
			
			if (!rcpt (sock, buff, NULL)) {
				g_free (buff);
				g_free (orig_rcpt);
				close (sock);
				return ERROR;
			}
			
			g_free (buff);
		}
		
		/* Do a RCPT TO for each of the carbon copies. */
		fields = get_num_fields (msg->cc_addr, ',');
		if (fields > 0) {
			for (i = 1; i <= fields; i++) {
				buff = get_field (msg->cc_addr, i, ',');
				
				if (!rcpt (sock, buff, orig_rcpt)) {
					g_free (buff);
					g_free (orig_rcpt);
					close (sock);
					return ERROR;
				}
				
				g_free (buff);
			}
		}
		
		/* Do a RCPT TO for each of the blind carbon copies. */
		fields = get_num_fields (msg->bcc_addr, ',');
		if (fields > 0) {
			for (i = 1; i <= fields; i++) {
				buff = get_field (msg->bcc_addr, i, ',');
				
				if (!rcpt (sock, buff, orig_rcpt)) {
					g_free (buff);
					g_free (orig_rcpt);
					close (sock);
					return ERROR;
				}
				
				g_free (buff);
			}
		}
		
		g_free (orig_rcpt);
	}
	
	if (!data (sock, msg->header, msg->body)) {
		/* we failed to send our data */
		close (sock);
		return ERROR;
	}
	
	if (!rset (sock)) {
		/* we failed to reset the smtp server */
		close (sock);
		return ERROR;
	}
	
	if (!quit (sock)) {
		/* we failed to quit - non-fatal */
		close (sock);
		return ERROR; /* should we return SUCCESS here? */
	}
	
	close (sock);                             /* close the socket */
	
	smtp_is_esmtp = FALSE;
	msg->status = msg->status | DELIVERED;   /* update status to delivered */
	
	return SUCCESS;                          /* everything went smoothly :) */
}
