/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*  Spruce
 *  Copyright (C) 1999-2000 Jeffrey Stedfast
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#ifdef HAVE_CONFIG_H
#	include <config.h>
#endif

#include "sort.h"

extern gint subject_sort_ignores_prefix;

gchar *strip_common_prefixes (gchar *string)
{
	gchar *pos;
	gint i;

	static gchar *common_prefixes[] = { " ", "\t", "Re:", "Fwd:", "Fw:", 
					    '\0' };

	if (subject_sort_ignores_prefix) {
		i = 0;
		pos = string;
		while (*pos && common_prefixes[i]) {
			if ((gchar *)strstrcase (pos, common_prefixes[i]) == pos) {
				pos += strlen (common_prefixes[i]);
				i = 0;
			} else {
				i++;
			}
		}
		return g_strdup (pos);
	} else {
		return g_strdup (string);
	}
}

gint gint_compare (gconstpointer a, gconstpointer b)
{
	gint first, second;

	g_return_val_if_fail(a, 0);
	g_return_val_if_fail(b, 0);

	first = GPOINTER_TO_INT(a);
	second = GPOINTER_TO_INT(b);

	if (first < second)
		return -1;
	if (first == second)
		return 0;
	return 1;
}

gint id_compare (GtkCList *clist, gconstpointer s1, gconstpointer s2)
{
	GtkCListRow *row1, *row2;
	gchar *text1, *text2;

	g_return_val_if_fail (clist != NULL, 0);
	g_return_val_if_fail (GTK_IS_CLIST (clist), 0);
	g_return_val_if_fail (s1 != NULL, 0);
	g_return_val_if_fail (s2 != NULL, 0);

	row1 = (GtkCListRow *) s1;
	row2 = (GtkCListRow *) s2;

	text1 = GTK_CELL_TEXT (row1->cell[clist->sort_column])->text;
	text2 = GTK_CELL_TEXT (row2->cell[clist->sort_column])->text;

	if (atoi(text1) < atoi(text2))
		return -1;
	if (atoi(text1) > atoi(text2))
		return 1;
	return 0;
}

gint str_compare (GtkCList *clist, gconstpointer s1, gconstpointer s2)
{
	/* Copyright (C) 1999 Jussi Junni */
	GtkCListRow *row1, *row2;
	gchar *text1, *text2;
	gint retval, text1length, text2length;

	g_return_val_if_fail (clist != NULL, 0);
	g_return_val_if_fail (GTK_IS_CLIST (clist), 0);
	g_return_val_if_fail (s1 != NULL, 0);
	g_return_val_if_fail (s2 != NULL, 0);

	row1 = (GtkCListRow *) s1;
	row2 = (GtkCListRow *) s2;

	text1 = strip_common_prefixes (GTK_CELL_TEXT (row1->cell[clist->sort_column])->text);
	text2 = strip_common_prefixes (GTK_CELL_TEXT (row2->cell[clist->sort_column])->text);
	/* Since we may be stripping a prefix, we may also need to compare
	   original text lengths, so the non-prefixed show up before the
	   prefixed (unless the responses/forwards don't match subjects, 
	   of course ;) */
	text1length = strlen (GTK_CELL_TEXT (row1->cell[clist->sort_column])->text);
	text2length = strlen (GTK_CELL_TEXT (row2->cell[clist->sort_column])->text);

	if (!text2) {
		retval = text1 != NULL;
	}
	else if (!text1) {
		retval = -1;
	} else {
		retval = g_strcasecmp (text1, text2);
		if (retval == 0) {
			if (text1length < text2length)
				retval = -1;
			else if (text1length > text2length)
				retval = 1;
			else
				retval = 0;
		}
	}

	g_free (text1);
	g_free (text2);
	return retval;
}

gint date_compare (GtkCList *clist, gconstpointer s1, gconstpointer s2)
{
	/* Original function by Jussi Junni, modified by Jeffrey Stedfast */
	GtkCListRow *row1, *row2;
	gchar *text1, *text2;
	date_t date1, date2;

	g_return_val_if_fail (clist != NULL, 0);
	g_return_val_if_fail (GTK_IS_CLIST (clist), 0);
	g_return_val_if_fail (s1 != NULL, 0);
	g_return_val_if_fail (s2 != NULL, 0);

	row1 = (GtkCListRow *) s1;
	row2 = (GtkCListRow *) s2;

	text1 = GTK_CELL_TEXT (row1->cell[clist->sort_column])->text;
	text2 = GTK_CELL_TEXT (row2->cell[clist->sort_column])->text;

	if (text2 == NULL)
		return (text1 != NULL);
	if (text1 == NULL)
		return -1;

	date1 = parse_date(text1);
	date2 = parse_date(text2);

	if (date1.year < date2.year) return 1;
	if (date1.year > date2.year) return -1;

	if (date1.month != 0 && date2.month != 0) {
		if (date1.month < date2.month) return 1;
		if (date1.month > date2.month) return -1;
	} else {
		fprintf(stderr, "Invalid date format! (%s)\n", date1.month == 0 ? text1 : text2);
		fflush(stderr);
	}

	if (date1.day < date2.day) return 1;
	if (date1.day > date2.day) return -1;

	if (date1.hour < date2.hour) return 1;
	if (date1.hour > date2.hour) return -1;

	if (date1.minute < date2.minute) return 1;
	if (date1.minute > date2.minute) return -1;

	if (date1.second < date2.second) return 1;
	if (date1.second > date2.second) return -1;
  
	return 0;
}
